@@ -47,12 +47,18 @@ static VALUE sGroup;
 #define HAVE_UNAME 1
 #endif
 
-#ifndef _WIN32
-char *getenv();
+#ifdef STDC_HEADERS
+# include <stdlib.h>
+#else
+# ifdef HAVE_STDLIB_H
+#  include <stdlib.h>
+# endif
 #endif
-char *getlogin();
+RUBY_EXTERN char *getlogin(void);
 
-#define RUBY_ETC_VERSION "1.4.0"
+#define RUBY_ETC_VERSION "1.4.5"
+
+#define SYMBOL_LIT(str) ID2SYM(rb_intern_const(str ""))
 
 #ifdef HAVE_RB_DEPRECATE_CONSTANT
 void rb_deprecate_constant(VALUE mod, const char *name);
@@ -62,6 +68,17 @@ void rb_deprecate_constant(VALUE mod, const char *name);
 
 #include "constdefs.h"
 
+#ifndef HAVE_RB_IO_DESCRIPTOR
+static int
+io_descriptor_fallback(VALUE io)
+{
+    rb_io_t *fptr;
+    GetOpenFile(io, fptr);
+    return fptr->fd;
+}
+#define rb_io_descriptor io_descriptor_fallback
+#endif
+
 #ifdef HAVE_RUBY_ATOMIC_H
 # include "ruby/atomic.h"
 #else
@@ -188,9 +205,10 @@ setup_passwd(struct passwd *pwd)
 #endif
 
 /* call-seq:
- *	getpwuid(uid)	->  Passwd
+ *	getpwuid(uid)	->  Etc::Passwd
  *
- * Returns the /etc/passwd information for the user with the given integer +uid+.
+ * Returns the <tt>/etc/passwd</tt> information for the user with the given
+ * integer +uid+.
  *
  * The information is returned as a Passwd struct.
  *
@@ -199,7 +217,7 @@ setup_passwd(struct passwd *pwd)
  *
  * See the unix manpage for <code>getpwuid(3)</code> for more detail.
  *
- * === Example:
+ * *Example:*
  *
  *	Etc.getpwuid(0)
  *	#=> #<struct Etc::Passwd name="root", passwd="x", uid=0, gid=0, gecos="root",dir="/root", shell="/bin/bash">
@@ -227,16 +245,16 @@ etc_getpwuid(int argc, VALUE *argv, VALUE obj)
 }
 
 /* call-seq:
- *	getpwnam(name)	->  Passwd
+ *	getpwnam(name)	->  Etc::Passwd
  *
- * Returns the /etc/passwd information for the user with specified login
- * +name+.
+ * Returns the <tt>/etc/passwd</tt> information for the user with specified
+ * login +name+.
  *
  * The information is returned as a Passwd struct.
  *
  * See the unix manpage for <code>getpwnam(3)</code> for more detail.
  *
- * === Example:
+ * *Example:*
  *
  *	Etc.getpwnam('root')
  *	#=> #<struct Etc::Passwd name="root", passwd="x", uid=0, gid=0, gecos="root",dir="/root", shell="/bin/bash">
@@ -291,17 +309,17 @@ each_passwd(void)
 #endif
 
 /* call-seq:
- *	Etc.passwd { |struct| block }	->  Passwd
- *	Etc.passwd			->  Passwd
+ *	passwd { |struct| block }
+ *	passwd				->  Etc::Passwd
  *
  * Provides a convenient Ruby iterator which executes a block for each entry
- * in the /etc/passwd file.
+ * in the <tt>/etc/passwd</tt> file.
  *
  * The code block is passed an Passwd struct.
  *
  * See ::getpwent above for details.
  *
- * Example:
+ * *Example:*
  *
  *     require 'etc'
  *
@@ -327,18 +345,19 @@ etc_passwd(VALUE obj)
 }
 
 /* call-seq:
- *	Etc::Passwd.each { |struct| block }	->  Passwd
+ *	Etc::Passwd.each { |struct| block }	->  Etc::Passwd
  *	Etc::Passwd.each			->  Enumerator
  *
- * Iterates for each entry in the /etc/passwd file if a block is given.
+ * Iterates for each entry in the <tt>/etc/passwd</tt> file if a block is
+ * given.
  *
  * If no block is given, returns the Enumerator.
  *
  * The code block is passed an Passwd struct.
  *
- * See ::getpwent above for details.
+ * See Etc.getpwent above for details.
  *
- * Example:
+ * *Example:*
  *
  *     require 'etc'
  *
@@ -360,8 +379,11 @@ etc_each_passwd(VALUE obj)
     return obj;
 }
 
-/* Resets the process of reading the /etc/passwd file, so that the next call
- * to ::getpwent will return the first entry again.
+/* call-seq:
+ *	setpwent
+ *
+ * Resets the process of reading the <tt>/etc/passwd</tt> file, so that the
+ * next call to ::getpwent will return the first entry again.
  */
 static VALUE
 etc_setpwent(VALUE obj)
@@ -372,8 +394,11 @@ etc_setpwent(VALUE obj)
     return Qnil;
 }
 
-/* Ends the process of scanning through the /etc/passwd file begun with
- * ::getpwent, and closes the file.
+/* call-seq:
+ *	endpwent
+ *
+ * Ends the process of scanning through the <tt>/etc/passwd</tt> file begun
+ * with ::getpwent, and closes the file.
  */
 static VALUE
 etc_endpwent(VALUE obj)
@@ -384,7 +409,10 @@ etc_endpwent(VALUE obj)
     return Qnil;
 }
 
-/* Returns an entry from the /etc/passwd file.
+/* call-seq:
+ *	getpwent	->  Etc::Passwd
+ *
+ * Returns an entry from the <tt>/etc/passwd</tt> file.
  *
  * The first time it is called it opens the file and returns the first entry;
  * each successive call returns the next entry, or +nil+ if the end of the file
@@ -432,16 +460,16 @@ setup_group(struct group *grp)
 #endif
 
 /* call-seq:
- *	getgrgid(group_id)  ->	Group
+ *	getgrgid(group_id)  ->	Etc::Group
  *
  * Returns information about the group with specified integer +group_id+,
- * as found in /etc/group.
+ * as found in <tt>/etc/group</tt>.
  *
  * The information is returned as a Group struct.
  *
  * See the unix manpage for <code>getgrgid(3)</code> for more detail.
  *
- * === Example:
+ * *Example:*
  *
  *	Etc.getgrgid(100)
  *	#=> #<struct Etc::Group name="users", passwd="x", gid=100, mem=["meta", "root"]>
@@ -470,16 +498,16 @@ etc_getgrgid(int argc, VALUE *argv, VALUE obj)
 }
 
 /* call-seq:
- *	getgrnam(name)	->  Group
+ *	getgrnam(name)	->  Etc::Group
  *
  * Returns information about the group with specified +name+, as found in
- * /etc/group.
+ * <tt>/etc/group</tt>.
  *
  * The information is returned as a Group struct.
  *
  * See the unix manpage for <code>getgrnam(3)</code> for more detail.
  *
- * === Example:
+ * *Example:*
  *
  *	Etc.getgrnam('users')
  *	#=> #<struct Etc::Group name="users", passwd="x", gid=100, mem=["meta", "root"]>
@@ -512,7 +540,6 @@ group_ensure(VALUE _)
     return Qnil;
 }
 
-
 static VALUE
 group_iterate(VALUE _)
 {
@@ -535,14 +562,18 @@ each_group(void)
 }
 #endif
 
-/* Provides a convenient Ruby iterator which executes a block for each entry
- * in the /etc/group file.
+/* call-seq:
+ *	group { |struct| block }
+ *	group				->  Etc::Group
+ *
+ * Provides a convenient Ruby iterator which executes a block for each entry
+ * in the <tt>/etc/group</tt> file.
  *
  * The code block is passed an Group struct.
  *
  * See ::getgrent above for details.
  *
- * Example:
+ * *Example:*
  *
  *     require 'etc'
  *
@@ -569,16 +600,17 @@ etc_group(VALUE obj)
 
 #ifdef HAVE_GETGRENT
 /* call-seq:
- *	Etc::Group.each { |group| block }   ->	obj
+ *	Etc::Group.each { |group| block }   ->	Etc::Group
  *	Etc::Group.each			    ->	Enumerator
  *
- * Iterates for each entry in the /etc/group file if a block is given.
+ * Iterates for each entry in the <tt>/etc/group</tt> file if a block is
+ * given.
  *
  * If no block is given, returns the Enumerator.
  *
  * The code block is passed a Group struct.
  *
- * Example:
+ * *Example:*
  *
  *     require 'etc'
  *
@@ -599,8 +631,11 @@ etc_each_group(VALUE obj)
 }
 #endif
 
-/* Resets the process of reading the /etc/group file, so that the next call
- * to ::getgrent will return the first entry again.
+/* call-seq:
+ *	setgrent
+ *
+ * Resets the process of reading the <tt>/etc/group</tt> file, so that the
+ * next call to ::getgrent will return the first entry again.
  */
 static VALUE
 etc_setgrent(VALUE obj)
@@ -611,8 +646,11 @@ etc_setgrent(VALUE obj)
     return Qnil;
 }
 
-/* Ends the process of scanning through the /etc/group file begun by
- * ::getgrent, and closes the file.
+/* call-seq:
+ *	endgrent
+ *
+ * Ends the process of scanning through the <tt>/etc/group</tt> file begun
+ * by ::getgrent, and closes the file.
  */
 static VALUE
 etc_endgrent(VALUE obj)
@@ -623,7 +661,10 @@ etc_endgrent(VALUE obj)
     return Qnil;
 }
 
-/* Returns an entry from the /etc/group file.
+/* call-seq:
+ *	getgrent	->  Etc::Group
+ *
+ * Returns an entry from the <tt>/etc/group</tt> file.
  *
  * The first time it is called it opens the file and returns the first entry;
  * each successive call returns the next entry, or +nil+ if the end of the file
@@ -652,14 +693,28 @@ etc_getgrent(VALUE obj)
 VALUE rb_w32_special_folder(int type);
 UINT rb_w32_system_tmpdir(WCHAR *path, UINT len);
 VALUE rb_w32_conv_from_wchar(const WCHAR *wstr, rb_encoding *enc);
+#elif defined(LOAD_RELATIVE)
+static inline VALUE
+rbconfig(void)
+{
+    VALUE config;
+    rb_require("rbconfig");
+    config = rb_const_get(rb_path2class("RbConfig"), rb_intern_const("CONFIG"));
+    Check_Type(config, T_HASH);
+    return config;
+}
 #endif
 
-/*
+/* call-seq:
+ *	sysconfdir	->  String
+ *
  * Returns system configuration directory.
  *
- * This is typically "/etc", but is modified by the prefix used when Ruby was
- * compiled. For example, if Ruby is built and installed in /usr/local,
- * returns "/usr/local/etc" on other platforms than Windows.
+ * This is typically <code>"/etc"</code>, but is modified by the prefix used
+ * when Ruby was compiled. For example, if Ruby is built and installed in
+ * <tt>/usr/local</tt>, returns <code>"/usr/local/etc"</code> on other
+ * platforms than Windows.
+ *
  * On Windows, this always returns the directory provided by the system.
  */
 static VALUE
@@ -667,12 +722,16 @@ etc_sysconfdir(VALUE obj)
 {
 #ifdef _WIN32
     return rb_w32_special_folder(CSIDL_COMMON_APPDATA);
+#elif defined(LOAD_RELATIVE)
+    return rb_hash_aref(rbconfig(), rb_str_new_lit("sysconfdir"));
 #else
     return rb_filesystem_str_new_cstr(SYSCONFDIR);
 #endif
 }
 
-/*
+/* call-seq:
+ *	systmpdir	->  String
+ *
  * Returns system temporary directory; typically "/tmp".
  */
 static VALUE
@@ -716,13 +775,15 @@ etc_systmpdir(VALUE _)
 }
 
 #ifdef HAVE_UNAME
-/*
+/* call-seq:
+ *	uname	-> hash
+ *
  * Returns the system information obtained by uname system call.
  *
  * The return value is a hash which has 5 keys at least:
  *   :sysname, :nodename, :release, :version, :machine
  *
- * Example:
+ * *Example:*
  *
  *   require 'etc'
  *   require 'pp'
@@ -764,12 +825,12 @@ etc_uname(VALUE obj)
 	sysname = "Windows";
 	break;
     }
-    rb_hash_aset(result, ID2SYM(rb_intern("sysname")), rb_str_new_cstr(sysname));
+    rb_hash_aset(result, SYMBOL_LIT("sysname"), rb_str_new_cstr(sysname));
     release = rb_sprintf("%lu.%lu.%lu", v.dwMajorVersion, v.dwMinorVersion, v.dwBuildNumber);
-    rb_hash_aset(result, ID2SYM(rb_intern("release")), release);
+    rb_hash_aset(result, SYMBOL_LIT("release"), release);
     version = rb_sprintf("%s Version %"PRIsVALUE": %"PRIsVALUE, sysname, release,
 			 rb_w32_conv_from_wchar(v.szCSDVersion, rb_utf8_encoding()));
-    rb_hash_aset(result, ID2SYM(rb_intern("version")), version);
+    rb_hash_aset(result, SYMBOL_LIT("version"), version);
 
 # if defined _MSC_VER && _MSC_VER < 1300
 #   define GET_COMPUTER_NAME(ptr, plen) GetComputerNameW(ptr, plen)
@@ -783,7 +844,7 @@ etc_uname(VALUE obj)
     }
     ALLOCV_END(vbuf);
     if (NIL_P(nodename)) nodename = rb_str_new(0, 0);
-    rb_hash_aset(result, ID2SYM(rb_intern("nodename")), nodename);
+    rb_hash_aset(result, SYMBOL_LIT("nodename"), nodename);
 
 # ifndef PROCESSOR_ARCHITECTURE_AMD64
 #   define PROCESSOR_ARCHITECTURE_AMD64 9
@@ -807,7 +868,7 @@ etc_uname(VALUE obj)
 	break;
     }
 
-    rb_hash_aset(result, ID2SYM(rb_intern("machine")), rb_str_new_cstr(mach));
+    rb_hash_aset(result, SYMBOL_LIT("machine"), rb_str_new_cstr(mach));
 #else
     struct utsname u;
     int ret;
@@ -818,11 +879,11 @@ etc_uname(VALUE obj)
         rb_sys_fail("uname");
 
     result = rb_hash_new();
-    rb_hash_aset(result, ID2SYM(rb_intern("sysname")), rb_str_new_cstr(u.sysname));
-    rb_hash_aset(result, ID2SYM(rb_intern("nodename")), rb_str_new_cstr(u.nodename));
-    rb_hash_aset(result, ID2SYM(rb_intern("release")), rb_str_new_cstr(u.release));
-    rb_hash_aset(result, ID2SYM(rb_intern("version")), rb_str_new_cstr(u.version));
-    rb_hash_aset(result, ID2SYM(rb_intern("machine")), rb_str_new_cstr(u.machine));
+    rb_hash_aset(result, SYMBOL_LIT("sysname"), rb_str_new_cstr(u.sysname));
+    rb_hash_aset(result, SYMBOL_LIT("nodename"), rb_str_new_cstr(u.nodename));
+    rb_hash_aset(result, SYMBOL_LIT("release"), rb_str_new_cstr(u.release));
+    rb_hash_aset(result, SYMBOL_LIT("version"), rb_str_new_cstr(u.version));
+    rb_hash_aset(result, SYMBOL_LIT("machine"), rb_str_new_cstr(u.machine));
 #endif
 
     return result;
@@ -832,7 +893,9 @@ etc_uname(VALUE obj)
 #endif
 
 #ifdef HAVE_SYSCONF
-/*
+/* call-seq:
+ *	sysconf(name)	->  Integer
+ *
  * Returns system configuration variable using sysconf().
  *
  * _name_ should be a constant under <code>Etc</code> which begins with <code>SC_</code>.
@@ -866,7 +929,9 @@ etc_sysconf(VALUE obj, VALUE arg)
 #endif
 
 #ifdef HAVE_CONFSTR
-/*
+/* call-seq:
+ *	confstr(name)	->  String
+ *
  * Returns system configuration variable using confstr().
  *
  * _name_ should be a constant under <code>Etc</code> which begins with <code>CS_</code>.
@@ -913,7 +978,9 @@ etc_confstr(VALUE obj, VALUE arg)
 #endif
 
 #ifdef HAVE_FPATHCONF
-/*
+/* call-seq:
+ *	pathconf(name)	->  Integer
+ *
  * Returns pathname configuration variable using fpathconf().
  *
  * _name_ should be a constant under <code>Etc</code> which begins with <code>PC_</code>.
@@ -932,14 +999,11 @@ io_pathconf(VALUE io, VALUE arg)
 {
     int name;
     long ret;
-    rb_io_t *fptr;
 
     name = NUM2INT(arg);
 
-    GetOpenFile(io, fptr);
-
     errno = 0;
-    ret = fpathconf(fptr->fd, name);
+    ret = fpathconf(rb_io_descriptor(io), name);
     if (ret == -1) {
         if (errno == 0) /* no limit */
             return Qnil;
@@ -1008,7 +1072,9 @@ etc_nprocessors_affin(void)
 }
 #endif
 
-/*
+/* call-seq:
+ *	nprocessors	->  Integer
+ *
  * Returns the number of online processors.
  *
  * The result is intended as the number of processes to
@@ -1018,7 +1084,7 @@ etc_nprocessors_affin(void)
  * - sched_getaffinity(): Linux
  * - sysconf(_SC_NPROCESSORS_ONLN): GNU/Linux, NetBSD, FreeBSD, OpenBSD, DragonFly BSD, OpenIndiana, Mac OS X, AIX
  *
- * Example:
+ * *Example:*
  *
  *   require 'etc'
  *   p Etc.nprocessors #=> 4
@@ -1027,7 +1093,7 @@ etc_nprocessors_affin(void)
  * process is bound to specific cpus. This is intended for getting better
  * parallel processing.
  *
- * Example: (Linux)
+ * *Example:* (Linux)
  *
  *   linux$ taskset 0x3 ./ruby -retc -e "p Etc.nprocessors"  #=> 2
  *
@@ -1067,16 +1133,17 @@ etc_nprocessors(VALUE obj)
 
 /*
  * The Etc module provides access to information typically stored in
- * files in the /etc directory on Unix systems.
+ * files in the <tt>/etc</tt> directory on Unix systems.
  *
  * The information accessible consists of the information found in the
- * /etc/passwd and /etc/group files, plus information about the system's
- * temporary directory (/tmp) and configuration directory (/etc).
+ * <tt>/etc/passwd</tt> and <tt>/etc/group</tt> files, plus information
+ * about the system's temporary directory (<tt>/tmp</tt>) and configuration
+ * directory (<tt>/etc</tt>).
  *
  * The Etc module provides a more reliable way to access information about
  * the logged in user than environment variables such as +$USER+.
  *
- * == Example:
+ * *Example:*
  *
  *     require 'etc'
  *
@@ -1096,13 +1163,26 @@ Init_etc(void)
 {
     VALUE mEtc;
 
-    #ifdef HAVE_RB_EXT_RACTOR_SAFE
-	RB_EXT_RACTOR_SAFE(true);
-    #endif
     mEtc = rb_define_module("Etc");
+    /* The version */
     rb_define_const(mEtc, "VERSION", rb_str_new_cstr(RUBY_ETC_VERSION));
     init_constants(mEtc);
 
+    /* Ractor-safe methods */
+#ifdef HAVE_RB_EXT_RACTOR_SAFE
+    RB_EXT_RACTOR_SAFE(true);
+#endif
+    rb_define_module_function(mEtc, "systmpdir", etc_systmpdir, 0);
+    rb_define_module_function(mEtc, "uname", etc_uname, 0);
+    rb_define_module_function(mEtc, "sysconf", etc_sysconf, 1);
+    rb_define_module_function(mEtc, "confstr", etc_confstr, 1);
+    rb_define_method(rb_cIO, "pathconf", io_pathconf, 1);
+    rb_define_module_function(mEtc, "nprocessors", etc_nprocessors, 0);
+
+    /* Non-Ractor-safe methods, see https://bugs.ruby-lang.org/issues/21115 */
+#ifdef HAVE_RB_EXT_RACTOR_SAFE
+    RB_EXT_RACTOR_SAFE(false);
+#endif
     rb_define_module_function(mEtc, "getlogin", etc_getlogin, 0);
 
     rb_define_module_function(mEtc, "getpwuid", etc_getpwuid, -1);
@@ -1118,13 +1198,9 @@ Init_etc(void)
     rb_define_module_function(mEtc, "setgrent", etc_setgrent, 0);
     rb_define_module_function(mEtc, "endgrent", etc_endgrent, 0);
     rb_define_module_function(mEtc, "getgrent", etc_getgrent, 0);
+
+    /* Uses RbConfig::CONFIG so does not work in a Ractor */
     rb_define_module_function(mEtc, "sysconfdir", etc_sysconfdir, 0);
-    rb_define_module_function(mEtc, "systmpdir", etc_systmpdir, 0);
-    rb_define_module_function(mEtc, "uname", etc_uname, 0);
-    rb_define_module_function(mEtc, "sysconf", etc_sysconf, 1);
-    rb_define_module_function(mEtc, "confstr", etc_confstr, 1);
-    rb_define_method(rb_cIO, "pathconf", io_pathconf, 1);
-    rb_define_module_function(mEtc, "nprocessors", etc_nprocessors, 0);
 
     sPasswd =  rb_struct_define_under(mEtc, "Passwd",
 				      "name",
@@ -1159,14 +1235,17 @@ Init_etc(void)
 				      NULL);
 #if 0
     /*
-     * Passwd is a Struct that contains the following members:
+     * Passwd is a placeholder Struct for user database on Unix systems.
+     *
+     * === The struct contains the following members
      *
      * name::
      *	    contains the short login name of the user as a String.
      * passwd::
      *	    contains the encrypted password of the user as a String.
-     *	    an 'x' is returned if shadow passwords are in use. An '*' is returned
-     *      if the user cannot log in using a password.
+     *	    an <code>'x'</code> is returned if shadow passwords are in
+     *	    use. An <code>'*'</code> is returned if the user cannot
+     *	    log in using a password.
      * uid::
      *	    contains the integer user ID (uid) of the user.
      * gid::
@@ -1176,25 +1255,24 @@ Init_etc(void)
      * shell::
      *	    contains the path to the login shell of the user as a String.
      *
-     * === The following members below are optional, and must be compiled with special flags:
+     * === The following members below are system-dependent
      *
      * gecos::
      *     contains a longer String description of the user, such as
      *	   a full name. Some Unix systems provide structured information in the
      *     gecos field, but this is system-dependent.
-     *     must be compiled with +HAVE_STRUCT_PASSWD_PW_GECOS+
      * change::
-     *     password change time(integer) must be compiled with +HAVE_STRUCT_PASSWD_PW_CHANGE+
+     *     password change time(integer).
      * quota::
-     *     quota value(integer) must be compiled with +HAVE_STRUCT_PASSWD_PW_QUOTA+
+     *     quota value(integer).
      * age::
-     *     password age(integer) must be compiled with +HAVE_STRUCT_PASSWD_PW_AGE+
+     *     password age(integer).
      * class::
-     *     user access class(string) must be compiled with +HAVE_STRUCT_PASSWD_PW_CLASS+
+     *     user access class(string).
      * comment::
-     *     comment(string) must be compiled with +HAVE_STRUCT_PASSWD_PW_COMMENT+
+     *     comment(string).
      * expire::
-     *	    account expiration time(integer) must be compiled with +HAVE_STRUCT_PASSWD_PW_EXPIRE+
+     *	    account expiration time(integer).
      */
     sPasswd = rb_define_class_under(mEtc, "Passwd", rb_cStruct);
 #endif
@@ -1210,9 +1288,9 @@ Init_etc(void)
 
 #if 0
     /*
-     * Group is a Struct that is only available when compiled with +HAVE_GETGRENT+.
+     * Group is a placeholder Struct for user group database on Unix systems.
      *
-     * The struct contains the following members:
+     * === The struct contains the following members
      *
      * name::
      *	    contains the name of the group as a String.
@@ -1221,7 +1299,7 @@ Init_etc(void)
      *	    returned if password access to the group is not available; an empty
      *	    string is returned if no password is needed to obtain membership of
      *	    the group.
-     *	    Must be compiled with +HAVE_STRUCT_GROUP_GR_PASSWD+.
+     *	    This is system-dependent.
      * gid::
      *	    contains the group's numeric ID as an integer.
      * mem::
@@ -1233,4 +1311,8 @@ Init_etc(void)
     rb_extend_object(sGroup, rb_mEnumerable);
     rb_define_singleton_method(sGroup, "each", etc_each_group, 0);
 #endif
+
+#if defined(HAVE_GETPWENT) || defined(HAVE_GETGRENT)
+    (void)safe_setup_str;
+#endif
 }
