@@ -249,6 +249,8 @@ module Test
     end
 
     module Parallel # :nodoc: all
+      attr_accessor :prefix
+
       def process_args(args = [])
         return @options if @options
         options = super
@@ -282,7 +284,7 @@ module Test
             options[:parallel] ||= 256 # number of tokens to acquire first
           end
         end
-        @worker_timeout = EnvUtil.apply_timeout_scale(options[:worker_timeout] || 180)
+        @worker_timeout = EnvUtil.apply_timeout_scale(options[:worker_timeout] || 1200)
         super
       end
 
@@ -298,7 +300,7 @@ module Test
 
         opts.separator "parallel test options:"
 
-        options[:retry] = true
+        options[:retry] = false
 
         opts.on '-j N', '--jobs N', /\A(t)?(\d+)\z/, "Allow run tests with N jobs at once" do |_, t, a|
           options[:testing] = true & t # For testing
@@ -370,8 +372,12 @@ module Test
           @io.puts(*args)
         end
 
-        def run(task,type)
-          @file = File.basename(task, ".rb")
+        def run(task, type, base = nil)
+          if base
+            @file = task.delete_prefix(base).chomp(".rb")
+          else
+            @file = File.basename(task, ".rb")
+          end
           @real_file = task
           begin
             puts "loadpath #{[Marshal.dump($:-@loadpath)].pack("m0")}"
@@ -406,16 +412,18 @@ module Test
         rescue IOError
         end
 
-        def quit
+        def quit(reason = :normal)
           return if @io.closed?
           @quit_called = true
-          @io.puts "quit"
+          @io.puts "quit #{reason}"
         rescue Errno::EPIPE => e
           warn "#{@pid}:#{@status.to_s.ljust(7)}:#{@file}: #{e.message}"
         end
 
         def kill
-          Process.kill(:KILL, @pid)
+          signal = RUBY_PLATFORM =~ /mswin|mingw/ ? :KILL : :SEGV
+          Process.kill(signal, @pid)
+          warn "worker #{to_s} does not respond; #{signal} is sent"
         rescue Errno::ESRCH
         end
 
@@ -533,15 +541,15 @@ module Test
         @workers.reject! do |worker|
           next unless cond&.call(worker)
           begin
-            Timeout.timeout(1) do
-              worker.quit
+            Timeout.timeout(5) do
+              worker.quit(cond ? :timeout : :normal)
             end
           rescue Errno::EPIPE
           rescue Timeout::Error
           end
           closed&.push worker
           begin
-            Timeout.timeout(0.2) do
+            Timeout.timeout(1) do
               worker.close
             end
           rescue Timeout::Error
@@ -554,7 +562,7 @@ module Test
         return if (closed ||= @workers).empty?
         pids = closed.map(&:pid)
         begin
-          Timeout.timeout(0.2 * closed.size) do
+          Timeout.timeout(1 * closed.size) do
             Process.waitall
           end
         rescue Timeout::Error
@@ -595,7 +603,7 @@ module Test
             worker.quit
             worker = launch_worker
           end
-          worker.run(task, type)
+          worker.run(task, type, (@prefix unless @options[:job_status] == :replace))
           @test_count += 1
 
           jobs_status(worker)
@@ -1450,9 +1458,8 @@ module Test
       def setup_options(opts, options)
         super
         opts.on_tail '--launchable-test-reports=PATH', String, 'Report test results in Launchable JSON format' do |path|
-          require 'json'
-          require 'uri'
-          options[:launchable_test_reports] = writer = JsonStreamWriter.new(path)
+          require_relative '../launchable'
+          options[:launchable_test_reports] = writer = Launchable::JsonStreamWriter.new(path)
           writer.write_array('testCases')
           main_pid = Process.pid
           at_exit {
@@ -1469,92 +1476,6 @@ module Test
           component.to_s.gsub('%', '%25').gsub('=', '%3D').gsub('#', '%23').gsub('&', '%26')
         end
       end
-
-      ##
-      # JsonStreamWriter writes a JSON file using a stream.
-      # By utilizing a stream, we can minimize memory usage, especially for large files.
-      class JsonStreamWriter
-        def initialize(path)
-          @file = File.open(path, "w")
-          @file.write("{")
-          @indent_level = 0
-          @is_first_key_val = true
-          @is_first_obj = true
-          write_new_line
-        end
-
-        def write_object obj
-          if @is_first_obj
-            @is_first_obj = false
-          else
-            write_comma
-            write_new_line
-          end
-          @indent_level += 1
-          @file.write(to_json_str(obj))
-          @indent_level -= 1
-          @is_first_key_val = true
-          # Occasionally, invalid JSON will be created as shown below, especially when `--repeat-count` is specified.
-          # {
-          #   "testPath": "file=test%2Ftest_timeout.rb&class=TestTimeout&testcase=test_allows_zero_seconds",
-          #   "status": "TEST_PASSED",
-          #   "duration": 2.7e-05,
-          #   "createdAt": "2024-02-09 12:21:07 +0000",
-          #   "stderr": null,
-          #   "stdout": null
-          # }: null <- here
-          # },
-          # To prevent this, IO#flush is called here.
-          @file.flush
-        end
-
-        def write_array(key)
-          @indent_level += 1
-          @file.write(to_json_str(key))
-          write_colon
-          @file.write(" ", "[")
-          write_new_line
-        end
-
-        def close
-          return if @file.closed?
-          close_array
-          @indent_level -= 1
-          write_new_line
-          @file.write("}", "\n")
-          @file.flush
-          @file.close
-        end
-
-        private
-        def to_json_str(obj)
-          json = JSON.pretty_generate(obj)
-          json.gsub(/^/, ' ' * (2 * @indent_level))
-        end
-
-        def write_indent
-          @file.write(" " * 2 * @indent_level)
-        end
-
-        def write_new_line
-          @file.write("\n")
-        end
-
-        def write_comma
-          @file.write(',')
-        end
-
-        def write_colon
-          @file.write(":")
-        end
-
-        def close_array
-          write_new_line
-          write_indent
-          @file.write("]")
-          @indent_level -= 1
-        end
-      end
     end
 
     class Runner # :nodoc: all
@@ -1761,9 +1682,7 @@ module Test
           puts if @verbose
           $stdout.flush
 
-          unless defined?(RubyVM::RJIT) && RubyVM::RJIT.enabled? # compiler process is wrongly considered as leak
-            leakchecker.check("#{inst.class}\##{inst.__name__}")
-          end
+          leakchecker.check("#{inst.class}\##{inst.__name__}")
 
           _end_method(inst)
 
@@ -1943,6 +1862,7 @@ module Test
         @force_standalone = force_standalone
         @runner = Runner.new do |files, options|
           base = options[:base_directory] ||= default_dir
+          @runner.prefix = base ? (base + "/") : nil
           files << default_dir if files.empty? and default_dir
           @to_run = files
           yield self if block_given?
