@@ -215,7 +215,12 @@ vm_call0_body(rb_execution_context_t *ec, struct rb_calling_info *calling, const
                 *reg_cfp->sp++ = argv[i];
             }
 
-            vm_call_iseq_setup(ec, reg_cfp, calling);
+            if (ISEQ_BODY(def_iseq_ptr(vm_cc_cme(cc)->def))->param.flags.forwardable) {
+                vm_call_iseq_fwd_setup(ec, reg_cfp, calling);
+            }
+            else {
+                vm_call_iseq_setup(ec, reg_cfp, calling);
+            }
             VM_ENV_FLAGS_SET(ec->cfp->ep, VM_FRAME_FLAG_FINISH);
             return vm_exec(ec); // CHECK_INTS in this function
         }
@@ -721,13 +726,6 @@ rb_check_funcall_with_hook_kw(VALUE recv, ID mid, int argc, const VALUE *argv,
     return rb_vm_call_kw(ec, recv, mid, argc, argv, me, kw_splat);
 }
 
-VALUE
-rb_check_funcall_with_hook(VALUE recv, ID mid, int argc, const VALUE *argv,
-                           rb_check_funcall_hook *hook, VALUE arg)
-{
-    return rb_check_funcall_with_hook_kw(recv, mid, argc, argv, hook, arg, RB_NO_KEYWORDS);
-}
-
 const char *
 rb_type_str(enum ruby_value_type type)
 {
@@ -1553,6 +1551,37 @@ rb_block_call_kw(VALUE obj, ID mid, int argc, const VALUE * argv,
     return rb_iterate_internal(iterate_method, (VALUE)&arg, bl_proc, data2);
 }
 
+/*
+ * A flexible variant of rb_block_call and rb_block_call_kw.
+ * This function accepts flags:
+ *
+ *   RB_NO_KEYWORDS, RB_PASS_KEYWORDS, RB_PASS_CALLED_KEYWORDS:
+ *   Works as the same as rb_block_call_kw.
+ *
+ *   RB_BLOCK_NO_USE_PACKED_ARGS:
+ *   The given block ("bl_proc") does not use "yielded_arg" of rb_block_call_func_t.
+ *   Instead, the block accesses the yielded arguments via "argc" and "argv".
+ *   This flag allows the called method to yield arguments without allocating an Array.
+ */
+VALUE
+rb_block_call2(VALUE obj, ID mid, int argc, const VALUE *argv,
+               rb_block_call_func_t bl_proc, VALUE data2, long flags)
+{
+    struct iter_method_arg arg;
+
+    arg.obj = obj;
+    arg.mid = mid;
+    arg.argc = argc;
+    arg.argv = argv;
+    arg.kw_splat = flags & 1;
+
+    struct vm_ifunc *ifunc = rb_vm_ifunc_proc_new(bl_proc, (void *)data2);
+    if (flags & RB_BLOCK_NO_USE_PACKED_ARGS)
+        ifunc->flags |= IFUNC_YIELD_OPTIMIZABLE;
+
+    return rb_iterate0(iterate_method, (VALUE)&arg, ifunc, GET_EC());
+}
+
 VALUE
 rb_lambda_call(VALUE obj, ID mid, int argc, const VALUE *argv,
                rb_block_call_func_t bl_proc, int min_argc, int max_argc,
@@ -1629,6 +1658,10 @@ pm_eval_make_iseq(VALUE src, VALUE fname, int line,
     const rb_iseq_t *const parent = vm_block_iseq(base_block);
     const rb_iseq_t *iseq = parent;
     VALUE name = rb_fstring_lit("<compiled>");
+
+    // Conditionally enable coverage depending on the current mode:
+    int coverage_enabled = ((rb_get_coverage_mode() & COVERAGE_TARGET_EVAL) != 0) ? 1 : 0;
+
     if (!fname) {
         fname = rb_source_location(&line);
     }
@@ -1638,16 +1671,18 @@ pm_eval_make_iseq(VALUE src, VALUE fname, int line,
     }
     else {
         fname = get_eval_default_path();
+        coverage_enabled = 0;
     }
 
     pm_parse_result_t result = { 0 };
     pm_options_line_set(&result.options, line);
+    result.node.coverage_enabled = coverage_enabled;
 
     // Cout scopes, one for each parent iseq, plus one for our local scope
     int scopes_count = 0;
     do {
         scopes_count++;
-    } while ((iseq = ISEQ_BODY(iseq)->parent_iseq) && (ISEQ_BODY(iseq)->type != ISEQ_TYPE_TOP));
+    } while ((iseq = ISEQ_BODY(iseq)->parent_iseq));
     pm_options_scopes_init(&result.options, scopes_count + 1);
 
     // Walk over the scope tree, adding known locals at the correct depths. The
@@ -1655,17 +1690,33 @@ pm_eval_make_iseq(VALUE src, VALUE fname, int line,
     // scopes array refer to root nodes on the tree, and higher indexes are the
     // leaf nodes.
     iseq = parent;
+    rb_encoding *encoding = rb_enc_get(src);
+
+#define FORWARDING_POSITIONALS_CHR '*'
+#define FORWARDING_POSITIONALS_STR "*"
+#define FORWARDING_KEYWORDS_CHR ':'
+#define FORWARDING_KEYWORDS_STR ":"
+#define FORWARDING_BLOCK_CHR '&'
+#define FORWARDING_BLOCK_STR "&"
+#define FORWARDING_ALL_CHR '.'
+#define FORWARDING_ALL_STR "."
+
     for (int scopes_index = 0; scopes_index < scopes_count; scopes_index++) {
+        VALUE iseq_value = (VALUE)iseq;
         int locals_count = ISEQ_BODY(iseq)->local_table_size;
+
         pm_options_scope_t *options_scope = &result.options.scopes[scopes_count - scopes_index - 1];
         pm_options_scope_init(options_scope, locals_count);
 
+        uint8_t forwarding = PM_OPTIONS_SCOPE_FORWARDING_NONE;
+
         for (int local_index = 0; local_index < locals_count; local_index++) {
             pm_string_t *scope_local = &options_scope->locals[local_index];
             ID local = ISEQ_BODY(iseq)->local_table[local_index];
 
             if (rb_is_local_id(local)) {
-                const char *name = rb_id2name(local);
+                VALUE name_obj = rb_id2str(local);
+                const char *name = RSTRING_PTR(name_obj);
                 size_t length = strlen(name);
 
                 // Explicitly skip numbered parameters. These should not be sent
@@ -1674,17 +1725,53 @@ pm_eval_make_iseq(VALUE src, VALUE fname, int line,
                     continue;
                 }
 
-                pm_string_constant_init(scope_local, name, strlen(name));
+                // Check here if this local can be represented validly in the
+                // encoding of the source string. If it _cannot_, then it should
+                // not be added to the constant pool as it would not be able to
+                // be referenced anyway.
+                if (rb_enc_str_coderange_scan(name_obj, encoding) == ENC_CODERANGE_BROKEN) {
+                    continue;
+                }
+
+                /* We need to duplicate the string because the Ruby string may
+                 * be embedded so compaction could move the string and the pointer
+                 * will change. */
+                char *name_dup = xmalloc(length + 1);
+                strlcpy(name_dup, name, length + 1);
+
+                RB_GC_GUARD(name_obj);
+
+                pm_string_owned_init(scope_local, (uint8_t *) name_dup, length);
+            } else if (local == idMULT) {
+                forwarding |= PM_OPTIONS_SCOPE_FORWARDING_POSITIONALS;
+                pm_string_constant_init(scope_local, FORWARDING_POSITIONALS_STR, 1);
+            } else if (local == idPow) {
+                forwarding |= PM_OPTIONS_SCOPE_FORWARDING_KEYWORDS;
+                pm_string_constant_init(scope_local, FORWARDING_KEYWORDS_STR, 1);
+            } else if (local == idAnd) {
+                forwarding |= PM_OPTIONS_SCOPE_FORWARDING_BLOCK;
+                pm_string_constant_init(scope_local, FORWARDING_BLOCK_STR, 1);
+            } else if (local == idDot3) {
+                forwarding |= PM_OPTIONS_SCOPE_FORWARDING_ALL;
+                pm_string_constant_init(scope_local, FORWARDING_ALL_STR, 1);
             }
         }
 
+        pm_options_scope_forwarding_set(options_scope, forwarding);
         iseq = ISEQ_BODY(iseq)->parent_iseq;
+
+        /* We need to GC guard the iseq because the code above malloc memory
+         * which could trigger a GC. Since we only use ISEQ_BODY, the compiler
+         * may optimize out the iseq local variable so we need to GC guard it. */
+        RB_GC_GUARD(iseq_value);
     }
 
     // Add our empty local scope at the very end of the array for our eval
     // scope's locals.
     pm_options_scope_init(&result.options.scopes[scopes_count], 0);
-    VALUE error = pm_parse_string(&result, src, fname);
+
+    VALUE script_lines;
+    VALUE error = pm_parse_string(&result, src, fname, ruby_vm_keep_script_lines ? &script_lines : NULL);
 
     // If the parse failed, clean up and raise.
     if (error != Qnil) {
@@ -1703,6 +1790,7 @@ pm_eval_make_iseq(VALUE src, VALUE fname, int line,
         RUBY_ASSERT(parent_scope != NULL);
 
         pm_options_scope_t *options_scope = &result.options.scopes[scopes_count - scopes_index - 1];
+        parent_scope->coverage_enabled = coverage_enabled;
         parent_scope->parser = &result.parser;
         parent_scope->index_lookup_table = st_init_numtable();
 
@@ -1712,14 +1800,38 @@ pm_eval_make_iseq(VALUE src, VALUE fname, int line,
 
         for (int local_index = 0; local_index < locals_count; local_index++) {
             const pm_string_t *scope_local = &options_scope->locals[local_index];
-
             pm_constant_id_t constant_id = 0;
-            if (pm_string_length(scope_local) > 0) {
-                constant_id = pm_constant_pool_insert_constant(
-                        &result.parser.constant_pool, pm_string_source(scope_local),
-                        pm_string_length(scope_local));
-                st_insert(parent_scope->index_lookup_table, (st_data_t)constant_id, (st_data_t)local_index);
+
+            const uint8_t *source = pm_string_source(scope_local);
+            size_t length = pm_string_length(scope_local);
+
+            if (length > 0) {
+                if (length == 1) {
+                    switch (*source) {
+                      case FORWARDING_POSITIONALS_CHR:
+                        constant_id = PM_CONSTANT_MULT;
+                        break;
+                      case FORWARDING_KEYWORDS_CHR:
+                        constant_id = PM_CONSTANT_POW;
+                        break;
+                      case FORWARDING_BLOCK_CHR:
+                        constant_id = PM_CONSTANT_AND;
+                        break;
+                      case FORWARDING_ALL_CHR:
+                        constant_id = PM_CONSTANT_DOT3;
+                        break;
+                      default:
+                        constant_id = pm_constant_pool_insert_constant(&result.parser.constant_pool, source, length);
+                        break;
+                    }
+                }
+                else {
+                    constant_id = pm_constant_pool_insert_constant(&result.parser.constant_pool, source, length);
+                }
+
+                st_insert(parent_scope->index_lookup_table, (st_data_t) constant_id, (st_data_t) local_index);
             }
+
             pm_constant_id_list_append(&parent_scope->locals, constant_id);
         }
 
@@ -1728,16 +1840,35 @@ pm_eval_make_iseq(VALUE src, VALUE fname, int line,
         iseq = ISEQ_BODY(iseq)->parent_iseq;
     }
 
-    iseq = pm_iseq_new_eval(&result.node, name, fname, Qnil, line, parent, 0);
+#undef FORWARDING_POSITIONALS_CHR
+#undef FORWARDING_POSITIONALS_STR
+#undef FORWARDING_KEYWORDS_CHR
+#undef FORWARDING_KEYWORDS_STR
+#undef FORWARDING_BLOCK_CHR
+#undef FORWARDING_BLOCK_STR
+#undef FORWARDING_ALL_CHR
+#undef FORWARDING_ALL_STR
+
+    int error_state;
+    iseq = pm_iseq_new_eval(&result.node, name, fname, Qnil, line, parent, 0, &error_state);
 
     pm_scope_node_t *prev = result.node.previous;
     while (prev) {
         pm_scope_node_t *next = prev->previous;
+        pm_constant_id_list_free(&prev->locals);
+        pm_scope_node_destroy(prev);
         ruby_xfree(prev);
         prev = next;
     }
 
     pm_parse_result_free(&result);
+
+    // If there was an error, raise it after memory has been cleaned up
+    if (error_state) {
+        RUBY_ASSERT(iseq == NULL);
+        rb_jump_tag(error_state);
+    }
+
     rb_exec_event_hook_script_compiled(GET_EC(), iseq, src);
 
     return iseq;
@@ -1747,7 +1878,7 @@ static const rb_iseq_t *
 eval_make_iseq(VALUE src, VALUE fname, int line,
                const struct rb_block *base_block)
 {
-    if (*rb_ruby_prism_ptr()) {
+    if (rb_ruby_prism_p()) {
         return pm_eval_make_iseq(src, fname, line, base_block);
     }
     const VALUE parser = rb_parser_new();
@@ -2530,11 +2661,31 @@ local_var_list_update(st_data_t *key, st_data_t *value, st_data_t arg, int exist
     return ST_CONTINUE;
 }
 
+extern int rb_numparam_id_p(ID id);
+
 static void
 local_var_list_add(const struct local_var_list *vars, ID lid)
 {
-    if (lid && rb_is_local_id(lid)) {
-        /* should skip temporary variable */
+    /* should skip temporary variable */
+    if (!lid) return;
+    if (!rb_is_local_id(lid)) return;
+
+    /* should skip numbered parameters as well */
+    if (rb_numparam_id_p(lid)) return;
+
+    st_data_t idx = 0;	/* tbl->num_entries */
+    rb_hash_stlike_update(vars->tbl, ID2SYM(lid), local_var_list_update, idx);
+}
+
+static void
+numparam_list_add(const struct local_var_list *vars, ID lid)
+{
+    /* should skip temporary variable */
+    if (!lid) return;
+    if (!rb_is_local_id(lid)) return;
+
+    /* should skip anything but numbered parameters */
+    if (rb_numparam_id_p(lid)) {
         st_data_t idx = 0;	/* tbl->num_entries */
         rb_hash_stlike_update(vars->tbl, ID2SYM(lid), local_var_list_update, idx);
     }
@@ -2662,6 +2813,17 @@ rb_current_realfilepath(void)
     return Qnil;
 }
 
+// Assert that an internal function is running and return
+// the imemo object that represents it.
+struct vm_ifunc *
+rb_current_ifunc(void)
+{
+    // Search VM_FRAME_MAGIC_IFUNC to see ifunc imemos put on the iseq field.
+    VALUE ifunc = (VALUE)GET_EC()->cfp->iseq;
+    RUBY_ASSERT_ALWAYS(imemo_type_p(ifunc, imemo_ifunc));
+    return (struct vm_ifunc *)ifunc;
+}
+
 void
 Init_vm_eval(void)
 {
