@@ -1,7 +1,6 @@
 //! Code to track assumptions made during code generation and invalidate
 //! generated code if and when these assumptions are invalidated.
 
-use crate::asm::OutlinedCb;
 use crate::backend::ir::Assembler;
 use crate::codegen::*;
 use crate::core::*;
@@ -31,7 +30,6 @@ pub struct Invariants {
     /// quick access to all of the blocks that are making this assumption when
     /// the operator is redefined.
     basic_operator_blocks: HashMap<(RedefinitionFlag, ruby_basic_operators), HashSet<BlockRef>>,
-
     /// A map from a block to a set of classes and their associated basic
     /// operators that the block is assuming are not redefined. This is used for
     /// quick access to all of the assumptions that a block is making when it
@@ -49,7 +47,6 @@ pub struct Invariants {
     /// a constant `A::B` is redefined, then all blocks that are assuming that
     /// `A` and `B` have not be redefined must be invalidated.
     constant_state_blocks: HashMap<ID, HashSet<BlockRef>>,
-
     /// A map from a block to a set of IDs that it is assuming have not been
     /// redefined.
     block_constant_states: HashMap<BlockRef, HashSet<ID>>,
@@ -58,6 +55,9 @@ pub struct Invariants {
     /// will have no singleton class. When the set is empty, it means that
     /// there has been a singleton class for the class after boot, so you cannot
     /// assume no singleton class going forward.
+    /// For now, the key can be only Array, Hash, or String. Consider making
+    /// an inverted HashMap if we start using this for user-defined classes
+    /// to maintain the performance of block_assumptions_free().
     no_singleton_classes: HashMap<VALUE, HashSet<BlockRef>>,
 
     /// A map from an ISEQ to a set of blocks that assume base pointer is equal
@@ -98,12 +98,11 @@ impl Invariants {
 pub fn assume_bop_not_redefined(
     jit: &mut JITState,
     asm: &mut Assembler,
-    ocb: &mut OutlinedCb,
     klass: RedefinitionFlag,
     bop: ruby_basic_operators,
 ) -> bool {
     if unsafe { BASIC_OP_UNREDEFINED_P(bop, klass) } {
-        if jit_ensure_block_entry_exit(jit, asm, ocb).is_none() {
+        if jit_ensure_block_entry_exit(jit, asm).is_none() {
             return false;
         }
         jit.bop_assumptions.push((klass, bop));
@@ -192,13 +191,12 @@ pub fn iseq_free_invariants(iseq: IseqPtr) {
 pub fn assume_method_basic_definition(
     jit: &mut JITState,
     asm: &mut Assembler,
-    ocb: &mut OutlinedCb,
     klass: VALUE,
     mid: ID
 ) -> bool {
     if unsafe { rb_method_basic_definition_p(klass, mid) } != 0 {
         let cme = unsafe { rb_callable_method_entry(klass, mid) };
-        jit.assume_method_lookup_stable(asm, ocb, cme);
+        jit.assume_method_lookup_stable(asm, cme);
         true
     } else {
         false
@@ -207,11 +205,11 @@ pub fn assume_method_basic_definition(
 
 /// Tracks that a block is assuming it is operating in single-ractor mode.
 #[must_use]
-pub fn assume_single_ractor_mode(jit: &mut JITState, asm: &mut Assembler, ocb: &mut OutlinedCb) -> bool {
+pub fn assume_single_ractor_mode(jit: &mut JITState, asm: &mut Assembler) -> bool {
     if unsafe { rb_yjit_multi_ractor_p() } {
         false
     } else {
-        if jit_ensure_block_entry_exit(jit, asm, ocb).is_none() {
+        if jit_ensure_block_entry_exit(jit, asm).is_none() {
             return false;
         }
         jit.block_assumes_single_ractor = true;
@@ -348,7 +346,7 @@ pub extern "C" fn rb_yjit_constant_state_changed(id: ID) {
 /// Callback for marking GC objects inside [Invariants].
 /// See `struct yjijt_root_struct` in C.
 #[no_mangle]
-pub extern "C" fn rb_yjit_root_mark(_: *mut c_void) {
+pub extern "C" fn rb_yjit_root_mark() {
     // Call rb_gc_mark on exit location's raw_samples to
     // wrap frames in a GC allocated object. This needs to be called
     // at the same time as root mark.
@@ -377,7 +375,7 @@ pub extern "C" fn rb_yjit_root_mark(_: *mut c_void) {
 }
 
 #[no_mangle]
-pub extern "C" fn rb_yjit_root_update_references(_: *mut c_void) {
+pub extern "C" fn rb_yjit_root_update_references() {
     if unsafe { INVARIANTS.is_none() } {
         return;
     }
@@ -465,11 +463,15 @@ pub fn block_assumptions_free(blockref: BlockRef) {
     }
 
     // Remove tracking for blocks assuming no singleton class
+    // NOTE: no_singleton_class has up to 3 keys (Array, Hash, or String) for now.
+    // This is effectively an O(1) access unless we start using it for more classes.
     for (_, blocks) in invariants.no_singleton_classes.iter_mut() {
         blocks.remove(&blockref);
     }
+
     // Remove tracking for blocks assuming EP doesn't escape
-    for (_, blocks) in invariants.no_ep_escape_iseqs.iter_mut() {
+    let iseq = unsafe { blockref.as_ref() }.get_blockid().iseq;
+    if let Some(blocks) = invariants.no_ep_escape_iseqs.get_mut(&iseq) {
         blocks.remove(&blockref);
     }
 }
@@ -624,6 +626,8 @@ pub extern "C" fn rb_yjit_tracing_invalidate_all() {
         return;
     }
 
+    incr_counter!(invalidate_everything);
+
     // Stop other ractors since we are going to patch machine code.
     with_vm_lock(src_loc!(), || {
         // Make it so all live block versions are no longer valid branch targets
@@ -686,7 +690,7 @@ pub extern "C" fn rb_yjit_tracing_invalidate_all() {
             cb.set_write_ptr(patch.inline_patch_pos);
             cb.set_dropped_bytes(false);
             cb.without_page_end_reserve(|cb| {
-                let mut asm = crate::backend::ir::Assembler::new();
+                let mut asm = crate::backend::ir::Assembler::new_without_iseq();
                 asm.jmp(patch.outlined_target_pos.as_side_exit());
                 if asm.compile(cb, None).is_none() {
                     panic!("Failed to apply patch at {:?}", patch.inline_patch_pos);
